from typing import List, Union

from banditbench.agents.classics import UCBAgent, LinUCBAgent
from banditbench.agents.llm import LLMAgent, LLMAgentBuilder
from banditbench.agents.guides import UCBGuide, LinUCBGuide
from banditbench.utils import calculate_cost

from banditbench.tasks import (create_small_gap_bernoulli_bandit, create_high_var_gaussian_bandit,
                               VerbalMultiArmedBandit, create_large_gap_bernoulli_bandit,
                               create_low_var_gaussian_bandit, MovieLens, MovieLensVerbal)


class BenchBase:
    def __init__(self, num_threads=10):
        self.n_trials = 20
        self.num_threads = num_threads
        self.envs, self.verbal_envs = self.create_envs()

    @property
    def name(self) -> str:
        raise NotImplementedError

    def create_envs(self):
        raise NotImplementedError

    def get_cost_eval_agent(self, env, verbal_env):
        raise NotImplementedError

    def evaluate(self, agents: List[LLMAgentBuilder], num_threads: int = None):
        """
        :param agents: LLMAgentBuilder is a list of partially initialized/configured agents, they are created through
                       LLMAgent.build() method (but did not pass in `env`)
        :return:
        """
        num_threads = self.num_threads if num_threads is None else num_threads
        env_to_agent_results = {}
        for i, env in enumerate(self.envs):
            assert hasattr(env, 'name'), "Environment must have a name attribute"
            env_to_agent_results[env.name] = {}

            for j, agent_builder in enumerate(agents):
                agent = agent_builder.build(env)
                assert hasattr(agent, 'name'), "Agent must have a name attribute"
                assert hasattr(agent, 'in_context_learn'), "Agent must have a in_context_learn method"
                print(f"Evaluating {agent.name} ({j}/{len(agents)}) on {env.name} ({i}/{len(self.envs)}) ")
                env_to_agent_results[env.name][agent.name] = agent.in_context_learn(env, n_trajs=self.n_trials,
                                                                                    num_threads=num_threads)

        return env_to_agent_results

    def calculate_eval_cost(self, model: Union[List[str], str] = 'gpt-3.5-turbo', verbose=False):
        # the way we calculate cost is by doing a data collection
        # we different Context Layer design: raw history, summarized history, summarized history + AG
        env_to_results = {}
        for i, (env, verbal_env) in enumerate(zip(self.envs, self.verbal_envs)):
            agent = self.get_cost_eval_agent(env, verbal_env)
            assert "oracle" in agent.name.lower(), "The agent must be an oracle agent to calculate cost"
            print(f"Calculating {env.name} ({i}/{len(self.envs)}) ")

            env_to_results[verbal_env.name] = agent.in_context_learn(verbal_env, n_trajs=1, num_threads=1)

        models = [model] if isinstance(model, str) else model
        model_to_env_costs = {}

        for model_name in models:
            env_to_cost = {}
            for env_name, data_buffer in env_to_results.items():
                data = data_buffer.to_sft_data()
                cost = 0
                for step in data[0]:
                    cost += calculate_cost(model_name, step['prompt'], step['label'])
                env_to_cost[env_name] = cost * self.n_trials
            model_to_env_costs[model_name] = env_to_cost

        # Pretty print costs per model and environment
        if verbose:
            for model_name, env_to_cost in model_to_env_costs.items():
                print(f"\nCost Analysis for {model_name}")
                print("-" * 50)
                print(f"{'Environment':<30} {'Cost ($)':<10}")
                print("-" * 50)

                total_cost = 0
                for env_name, cost in env_to_cost.items():
                    print(f"{env_name:<30} ${cost:.4f}")
                    total_cost += cost

                print("-" * 50)
                print(f"{'Total Cost':<30} ${total_cost:.4f}")
                print(f"\nEstimated total cost for {len(self.envs)} environments "
                      f"with {self.n_trials} trials: ${total_cost:.2f}")

        # Print final summary table with just model names and total costs
        print("\nFinal Cost Summary")
        print("-" * 30)
        print(f"{'Model':<20} {'Total Cost ($)':<10}")
        print("-" * 30)
        # Sort models by total cost
        sorted_models = sorted(model_to_env_costs.items(),
                               key=lambda x: sum(x[1].values()))
        for model_name, env_to_cost in sorted_models:
            total = sum(env_to_cost.values())
            print(f"{model_name:<20} ${total:.2f}")
        print("-" * 30)

        return model_to_env_costs


class MABBench:
    def get_cost_eval_agent(self, env, verbal_env) -> LLMAgent:
        oracle = UCBAgent(env)
        # ucb_guide = UCBGuide(env)
        # agent = LLMAgent.build_with_env(verbal_env, ucb_guide, oracle)
        agent = LLMAgent.build_with_env(verbal_env, oracle)  # raw history agent
        return agent


class CoreBench(MABBench, BenchBase):
    def __init__(self, num_threads=10):
        super().__init__(num_threads=num_threads)

    def create_envs(self):
        # we create 8 bernoulli tasks -- focus on the easiest and hardest settings
        # this can be used to test both small and large models (smaller models should be able to solve the easy tasks)
        eval_envs = []
        eval_verbal_envs = []
        exp_id = 1
        for num_arms in [20]:
            for domain in ['VideoWatching', 'ClothesShopping']:
                for gap in ['hard']:
                    bern_env = create_small_gap_bernoulli_bandit(num_arms, horizon=1000, seed=exp_id)
                    eval_envs.append(bern_env)
                    bern_verbal_env = VerbalMultiArmedBandit(bern_env, domain, scenario_seed=exp_id)
                    eval_verbal_envs.append(bern_verbal_env)

                    exp_id += 1

        for num_arms in [5]:
            for domain in ['VideoWatching', 'ClothesShopping']:
                for gap in ['easy']:
                    bern_env = create_small_gap_bernoulli_bandit(num_arms, horizon=1000, seed=exp_id)
                    eval_envs.append(bern_env)
                    bern_verbal_env = VerbalMultiArmedBandit(bern_env, domain, scenario_seed=exp_id)
                    eval_verbal_envs.append(bern_verbal_env)

                    exp_id += 1

        return eval_envs, eval_verbal_envs

    @property
    def name(self) -> str:
        return "CoreBench"


class HardCoreBench(MABBench, BenchBase):
    """
    If you include the highest performing model, you can use this set of tasks.
    It is sufficiently challenging, and does not cost too much to evaluate.
    1000 * 30 = 30000 calls for one model.

    We evaluate on a small set of tasks that are considered hard for the agent.
    This includes 4 MAB, Bernoulli, Hard case
    """

    def __init__(self, num_threads=10):
        super().__init__(num_threads=num_threads)

    def create_envs(self):
        # we create 4 bernoulli tasks, 4 gaussian tasks
        eval_envs = []
        eval_verbal_envs = []
        exp_id = 1
        for num_arms in [20]:
            for domain in ['VideoWatching', 'ClothesShopping']:
                for gap in ['hard']:
                    bern_env = create_small_gap_bernoulli_bandit(num_arms, horizon=1000, seed=exp_id)
                    eval_envs.append(bern_env)
                    bern_verbal_env = VerbalMultiArmedBandit(bern_env, domain, scenario_seed=exp_id)
                    eval_verbal_envs.append(bern_verbal_env)

                    exp_id += 1

        return eval_envs, eval_verbal_envs

    @property
    def name(self) -> str:
        return "HardCoreBench"


class HardCorePlusBench(MABBench, BenchBase):
    """
    In addition to hard instances, now models need to understand floating point numbers
    """

    def __init__(self, num_threads=10):
        super().__init__(num_threads=num_threads)

    def create_envs(self):
        # we create 4 bernoulli tasks
        eval_envs = []
        eval_verbal_envs = []
        exp_id = 1
        for num_arms in [20]:
            for domain in ['VideoWatching', 'ClothesShopping']:
                for gap in ['hard']:
                    bern_env = create_small_gap_bernoulli_bandit(num_arms, horizon=1000, seed=exp_id)
                    bern_verbal_env = VerbalMultiArmedBandit(bern_env, domain, scenario_seed=exp_id)
                    eval_envs.append(bern_env)
                    eval_verbal_envs.append(bern_verbal_env)

                    gauss_env = create_high_var_gaussian_bandit(num_arms, horizon=1000, seed=exp_id)
                    eval_envs.append(gauss_env)
                    gauss_verbal_env = VerbalMultiArmedBandit(gauss_env, domain, scenario_seed=exp_id)
                    eval_verbal_envs.append(gauss_verbal_env)

                    exp_id += 1

        return eval_envs, eval_verbal_envs

    @property
    def name(self) -> str:
        return "HardCorePlusBench"


class FullBench(MABBench, BenchBase):
    def __init__(self, num_threads=10):
        super().__init__(num_threads=num_threads)

    def create_envs(self):
        eval_envs = []
        eval_verbal_envs = []
        exp_id = 1
        for num_arms in [5, 20]:
            for domain in ['VideoWatching', 'ClothesShopping']:
                for gap in ['easy', 'hard']:
                    if gap == 'easy':
                        bern_env = create_large_gap_bernoulli_bandit(num_arms, 300, seed=exp_id)
                        gauss_env = create_low_var_gaussian_bandit(num_arms, 300, seed=exp_id)
                    else:
                        bern_env = create_small_gap_bernoulli_bandit(num_arms, 1000, seed=exp_id)
                        gauss_env = create_high_var_gaussian_bandit(num_arms, 1000, seed=exp_id)

                    eval_envs.append(bern_env)
                    eval_envs.append(gauss_env)

                    bern_verbal_env = VerbalMultiArmedBandit(bern_env, domain, scenario_seed=exp_id)
                    gauss_verbal_env = VerbalMultiArmedBandit(gauss_env, domain, scenario_seed=exp_id)

                    eval_verbal_envs.append(bern_verbal_env)
                    eval_verbal_envs.append(gauss_verbal_env)

                    exp_id += 1

        return eval_envs, eval_verbal_envs

    @property
    def name(self) -> str:
        return "FullBench"


class CBBench:
    def get_cost_eval_agent(self, env, verbal_env) -> LLMAgent:
        oracle = LinUCBAgent(env)
        ucb_guide = LinUCBGuide(env)
        agent = LLMAgent.build_with_env(verbal_env, ucb_guide, oracle)
        return agent


class MovieBench(CBBench, BenchBase):
    """
    Evaluate on Contextual Bandit
    """

    def __init__(self, num_threads=10, toy=False):
        """
        :param toy: MovieLens dataset takes a long time to load. If toy=False, we load from 1m-ratings, which is
                    the dataset used in the paper. If toy=True, we load from 100k-ratings, which is much faster to
                    debug with.
        """
        self.toy = toy
        super().__init__(num_threads=num_threads)

    def create_envs(self):
        eval_envs = []
        eval_verbal_envs = []

        exp_id = 1
        for num_arms in [10, 30]:
            task_name = '1m-ratings' if not self.toy else '100k-ratings'
            env = MovieLens(task_name, num_arms=num_arms, horizon=200, rank_k=20, mode='train',
                            save_data_dir='./tensorflow_datasets/', seed=exp_id)
            verbal_env = MovieLensVerbal(env)
            eval_envs.append(env)
            eval_verbal_envs.append(verbal_env)

        return eval_envs, eval_verbal_envs

    @property
    def name(self) -> str:
        return "MovieBench"



import re

from banditbench.tasks.mab import BernoulliBandit, VerbalMultiArmedBandit
from banditbench.tasks.cb.movielens import MovieLens, MovieLensVerbal

from banditbench.agents.guides import VerbalGuideBase, UCBGuide, LinUCBGuide
from banditbench.agents.classics import UCBAgent, LinUCBAgent

from banditbench.agents.llm import (LLMMABAgentRH, LLMMABAgentSH, LLMCBAgentRH, LLMCBAgentRHWithAG, LLMMABAgentSHWithAG,
                                    OracleLLMCBAgentRH, OracleLLMCBAgentRHWithAG, OracleLLMMABAgentSH,
                                    OracleLLMMAbAgentRH, OracleLLMMABAgentSHWithAG, LLMAgent)

core_bandit = BernoulliBandit(2, 10, [0.5, 0.2], 123)
verbal_bandit = VerbalMultiArmedBandit(core_bandit, "VideoWatching")

env = None
verbal_env = None


def init_cb_env():
    global env
    global verbal_env

    if env is None:
        env = MovieLens('100k-ratings', num_arms=5, horizon=200, rank_k=5, mode='train',
                        save_data_dir='./tensorflow_datasets/')
    if verbal_env is None:
        verbal_env = MovieLensVerbal(env)


def mab_fake_steps(agent):
    _, reward, done, info = verbal_bandit.step('A')

    action = info['interaction'].mapped_action
    agent.update(action, reward, info)

    _, reward, done, info = verbal_bandit.step('B')
    action = info['interaction'].mapped_action
    agent.update(action, reward, info)


def test_mab_agent_rh():
    agent = LLMMABAgentRH(verbal_bandit)
    agent.generate = lambda x: x  # so that the LLM is not triggered

    mab_fake_steps(agent)

    constructed_llm_message = agent.act()
    assert "A video" in constructed_llm_message
    print(constructed_llm_message)


def test_mab_agent_sh():
    agent = LLMMABAgentSH(verbal_bandit)
    agent.generate = lambda x: x  # so that the LLM is not triggered

    mab_fake_steps(agent)
    constructed_llm_message = agent.act()
    assert "A video" in constructed_llm_message
    assert "B video" in constructed_llm_message

    assert "1 times" in constructed_llm_message
    assert "average reward" in constructed_llm_message

    print(constructed_llm_message)


def extract_values(text):
    pattern = r'average reward ([\d.]+).*?exploitation value ([\d.]+)'
    matches = re.finditer(pattern, text)

    results = []
    for match in matches:
        avg_reward = float(match.group(1))
        exploit_value = float(match.group(2))
        results.append({
            'average_reward': avg_reward,
            'exploitation_value': exploit_value
        })
    return results


def equal_pulls_equal_bonus(text):
    # Extract all video data with regex
    pattern = r'(\w+) video, (\d+) times.*?exploration bonus ([\d.]+)'
    matches = re.finditer(pattern, text)

    # Store data in format: {num_pulls: [(video_name, exploration_bonus)]}
    pulls_data = {}

    # Parse and group by number of pulls
    for match in matches:
        video_name = match.group(1)
        num_pulls = int(match.group(2))
        exploration_bonus = float(match.group(3))

        if num_pulls not in pulls_data:
            pulls_data[num_pulls] = []
        pulls_data[num_pulls].append((video_name, exploration_bonus))

    # Check each group of videos with same pulls
    inconsistencies = []
    for pulls, videos in pulls_data.items():
        if len(videos) > 1:  # If there are multiple videos with same pulls
            first_bonus = videos[0][1]
            for video_name, bonus in videos[1:]:
                if bonus != first_bonus:
                    inconsistencies.append(
                        f"Inconsistency found: Videos with {pulls} pulls have different bonuses. "
                        f"{videos[0][0]} video: {videos[0][1]}, {video_name} video: {bonus}"
                    )

    # Return test results
    if not inconsistencies:
        return True, "All videos with equal pulls have equal exploration bonuses."
    else:
        return False, "\n".join(inconsistencies)


def test_mab_agent_sh_with_ag():
    guide = UCBGuide(UCBAgent(core_bandit))
    agent = LLMMABAgentSHWithAG(verbal_bandit, guide)
    agent.generate = lambda x: x  # so that the LLM is not triggered

    mab_fake_steps(agent)
    constructed_llm_message = agent.act()
    results = extract_values(constructed_llm_message)

    for result in results:
        assert result['average_reward'] == result['exploitation_value']

    pass_flag, message = equal_pulls_equal_bonus(constructed_llm_message)
    assert pass_flag, message
    print(constructed_llm_message)


def cb_fake_steps(agent):
    state, info = verbal_env.reset()

    new_state, reward, _, info = verbal_env.step(state, 'Star Wars (1977)')

    action = info['interaction'].mapped_action
    agent.update(state, action, reward, info)

    state = new_state

    new_state, reward, _, info = verbal_env.step(state, 'Fargo (1996)')
    action = info['interaction'].mapped_action
    agent.update(state, action, reward, info)

    state = new_state
    return state


def test_cb_agent_rh():
    init_cb_env()
    agent = LLMCBAgentRH(verbal_env)
    agent.generate = lambda x: x  # so that the LLM is not triggered

    state = cb_fake_steps(agent)

    constructed_llm_message = agent.act(state)
    print(constructed_llm_message)


def test_cb_agent_sh_with_ag():
    init_cb_env()
    guide = LinUCBGuide(LinUCBAgent(env))
    agent = LLMCBAgentRHWithAG(verbal_env, guide)
    agent.generate = lambda x: x  # so that the LLM is not triggered

    state = cb_fake_steps(agent)

    constructed_llm_message = agent.act(state)
    print(constructed_llm_message)

# need to test reset!!!!!
def test_mab_agent_rh_reset():
    agent = LLMMABAgentRH(verbal_bandit)
    agent.generate = lambda x: x  # so that the LLM is not triggered

    mab_fake_steps(agent)

    agent.reset()
    assert len(agent.interaction_history) == 0

def test_mab_agent_sh_reset():
    agent = LLMMABAgentSH(verbal_bandit)
    agent.generate = lambda x: x  # so that the LLM is not triggered

    mab_fake_steps(agent)
    agent.reset()

def test_mab_agent_sh_with_ag_reset():
    guide = UCBGuide(UCBAgent(core_bandit))
    agent = LLMMABAgentSHWithAG(verbal_bandit, guide)
    agent.generate = lambda x: x  # so that the LLM is not triggered

    mab_fake_steps(agent)

    agent.reset()

    # fully reset both agent and the AlgorithmGuide's agent
    assert len(agent.interaction_history) == 0
    assert agent.ag.agent.arms[0] == 0

def test_cb_agent_rh_reset():
    init_cb_env()
    agent = LLMCBAgentRH(verbal_env)
    agent.generate = lambda x: x  # so that the LLM is not triggered

    state = cb_fake_steps(agent)

    agent.reset()

    assert len(agent.interaction_history) == 0

def test_cb_agent_sh_with_ag_reset():
    init_cb_env()
    guide = LinUCBGuide(LinUCBAgent(env))
    agent = LLMCBAgentRHWithAG(verbal_env, guide)
    agent.generate = lambda x: x  # so that the LLM is not triggered

    state = cb_fake_steps(agent)

    agent.reset()

    assert len(agent.interaction_history) == 0
    assert len(agent.ag_info_history) == 0
    assert agent.ag.agent.b[0][0] == 0

def test_general_llm_agent_construction():
    # Test MAB agents
    agent = LLMAgent.build_with_env(verbal_bandit)
    assert isinstance(agent, LLMMABAgentRH)

    agent = LLMAgent.build_with_env(verbal_bandit, summary=True)
    assert isinstance(agent, LLMMABAgentSH)

    guide = UCBGuide(UCBAgent(core_bandit))
    agent = LLMAgent.build_with_env(verbal_bandit, guide)
    assert isinstance(agent, LLMMABAgentSHWithAG)

    oracle = UCBAgent(core_bandit)
    agent = LLMAgent.build_with_env(verbal_bandit, oracle)
    assert isinstance(agent, OracleLLMMAbAgentRH)

    agent = LLMAgent.build_with_env(verbal_bandit, guide, oracle)
    assert isinstance(agent, OracleLLMMABAgentSHWithAG)

    # Test CB agents
    init_cb_env()
    agent = LLMAgent.build_with_env(verbal_env)
    assert isinstance(agent, LLMCBAgentRH)

    guide = LinUCBGuide(LinUCBAgent(env))
    agent = LLMAgent.build_with_env(verbal_env, guide)
    assert isinstance(agent, LLMCBAgentRHWithAG)

    oracle = LinUCBAgent(env)
    agent = LLMAgent.build_with_env(verbal_env, oracle)
    assert isinstance(agent, OracleLLMCBAgentRH)

    agent = LLMAgent.build_with_env(verbal_env, guide, oracle)
    assert isinstance(agent, OracleLLMCBAgentRHWithAG)

def test_llm_agent_builder_construction():
    # Test MAB agents
    agent = LLMAgent.build()
    agent = agent.build_with_env(verbal_bandit)
    assert isinstance(agent, LLMMABAgentRH)

    agent = LLMAgent.build(summary=True)
    agent = agent.build_with_env(verbal_bandit)
    assert isinstance(agent, LLMMABAgentSH)

    guide = UCBGuide(core_bandit)
    agent = LLMAgent.build(guide, summary=True)
    agent = agent.build_with_env(verbal_bandit)
    assert isinstance(agent, LLMMABAgentSHWithAG)

    oracle = UCBAgent(core_bandit)
    agent = LLMAgent.build(oracle)
    agent = agent.build_with_env(verbal_bandit)
    assert isinstance(agent, OracleLLMMAbAgentRH)

    agent = LLMAgent.build(guide, oracle)
    agent = agent.build_with_env(verbal_bandit)
    assert isinstance(agent, OracleLLMMABAgentSHWithAG)


from banditbench.tasks.mab.env import BernoulliBandit
from banditbench.agents.classics import UCBAgent, ThompsonSamplingAgent


def test_ucb_agent():
    # construct it, and then print out the action guide
    core_bandit = BernoulliBandit(2, 10, [0.5, 0.2], 123)
    ucb_agent = UCBAgent(core_bandit)
    ucb_agent.reset()

    action = ucb_agent.select_arm()
    assert type(action) == int
    assert action < core_bandit.num_arms and action >= 0

    # print(ucb_agent.get_guide_info()[0].to_str())


def test_thompson_sampling_agent():
    core_bandit = BernoulliBandit(2, 10, [0.5, 0.2], 123)
    ts_agent = ThompsonSamplingAgent(core_bandit)
    ts_agent.reset()

    action = ts_agent.select_arm()
    assert type(action) == int
    assert action < core_bandit.num_arms and action >= 0

    # print(ts_agent.get_guide_info()[0].to_str())

test_ucb_agent()
test_thompson_sampling_agent()



from banditbench import HardCoreBench, HardCorePlusBench, FullBench, CoreBench, MovieBench

def test_calculate_core_bench_cost():
    bench = CoreBench()
    model_to_cost = {}
    cost = bench.calculate_eval_cost([
        'gemini-1.5-pro',
        'gemini-1.5-flash',
        'gpt-4o-2024-11-20',
        "gpt-4o-mini-2024-07-18",
        "o1-2024-12-17",
        "o1-mini-2024-09-12",
        "claude-3-5-sonnet-20241022",
        "claude-3-5-haiku-20241022",
        "deepseek/deepseek-chat",
        # "deepseek/deepseek-reasoner"
    ])


def test_calculate_hardcore_bench_cost():
    bench = HardCoreBench()
    model_to_cost = {}
    cost = bench.calculate_eval_cost([
        'gemini-1.5-pro',
        'gemini-1.5-flash',
        'gpt-4o-2024-11-20',
        "gpt-4o-mini-2024-07-18",
        "o1-2024-12-17",
        "o1-mini-2024-09-12",
        "claude-3-5-sonnet-20241022",
        "claude-3-5-haiku-20241022",
        "deepseek/deepseek-chat",
        # "deepseek/deepseek-reasoner"
    ])

def test_calculate_hardcore_plus_bench_cost():
    bench = HardCorePlusBench()
    model_to_cost = {}
    cost = bench.calculate_eval_cost([
        'gemini-1.5-pro',
        'gemini-1.5-flash',
        'gpt-4o-2024-11-20',
        "gpt-4o-mini-2024-07-18",
        "o1-2024-12-17",
        "o1-mini-2024-09-12",
        "claude-3-5-sonnet-20241022",
        "claude-3-5-haiku-20241022",
        "deepseek/deepseek-chat",
    ])

def test_calculate_full_bench_cost():
    bench = FullBench()
    model_to_cost = {}
    cost = bench.calculate_eval_cost([
        'gemini-1.5-pro',
        'gemini-1.5-flash',
        'gpt-4o-2024-11-20',
        "gpt-4o-mini-2024-07-18",
        "o1-2024-12-17",
        "o1-mini-2024-09-12",
        "claude-3-5-sonnet-20241022",
        "claude-3-5-haiku-20241022",
        "deepseek/deepseek-chat"
    ])

# def test_calculate_moviebench_cost():
#     bench = MovieBench(toy=True)
#     model_to_cost = {}
#     cost = bench.calculate_eval_cost([
#         'gemini-1.5-pro',
#         'gemini-1.5-flash',
#         'gpt-4o-2024-11-20',
#         "gpt-4o-mini-2024-07-18",
#         "o1-2024-12-17",
#         "o1-mini-2024-09-12",
#         "claude-3-5-sonnet-20241022",
#         "claude-3-5-haiku-20241022"
#     ])


"""
We have a rather elaborate random seed strategy, controlled in the sampling process.
We test whether given the same master random seed (set in the env),
Both LLM and UCB can get the same reward if they choose the same action.
"""

from banditbench.tasks.mab import BernoulliBandit, VerbalMultiArmedBandit
from banditbench.agents.llm import LLMAgent
from banditbench.agents.classics import UCBAgent

def test_compare_env():
    core_bandit = BernoulliBandit(5, horizon=2, arm_params=[0.2, 0.2, 0.2, 0.2, 0.5], seed=321)
    verbal_bandit = VerbalMultiArmedBandit(core_bandit, "VideoWatching")

    agent = LLMAgent.build_with_env(verbal_bandit, summary=True, model="gpt-3.5-turbo")
    agent.generate = lambda x: "A"

    data = agent.in_context_learn(verbal_bandit, 5)

    classic_agent = UCBAgent(core_bandit)
    data_from_ucb = classic_agent.in_context_learn(core_bandit, 5)

    initial_reward = []
    for i in range(3):
        print(data.trajectories[i][0])
        initial_reward.append(data.trajectories[i][0].reward)

    initial_reward_ucb = []
    for i in range(3):
        print(data_from_ucb.trajectories[i][0])
        initial_reward_ucb.append(data_from_ucb.trajectories[i][0].reward)

    # write an assertion
    assert initial_reward == initial_reward_ucb


import numpy as np
import scipy
import scipy.stats
import matplotlib.pyplot as plt

from typing import List, Dict
from litellm.cost_calculator import completion_cost

plt.rcParams.update(
    {
        "font.size": 13,  # Default font size for text
        "axes.titlesize": 16,  # Font size for axes titles
        "axes.labelsize": 14,  # Font size for axes labels
        "xtick.labelsize": 13,  # Font size for x-tick labels
        "ytick.labelsize": 13,  # Font size for y-tick labels
        "legend.fontsize": 13,  # Font size for legend
        "figure.titlesize": 16,  # Font size for figure title
        "axes.formatter.useoffset": False,
        "axes.formatter.offset_threshold": 1,
    }
)
plt.rcParams["font.family"] = "DeJavu Serif"
plt.rcParams["font.serif"] = ["Times New Roman"]


def compute_pad(total_elements, K, desired_pad=5):
    # equal spaced sampling for few-shot in-context learning
    available_space = total_elements - K
    pad = min(desired_pad, available_space)
    return max(0, pad)


def dedent(text: str):
    """
    Remove leading and trailing whitespace for each line
    For example:
        ```
        Line 1 has no leading space
            Line 2 has two leading spaces
        ```
        The output will be :
        ```
        Line 1 has no leading space
        Line 2 has two leading spaces
        ```
    This allows writing cleaner multiline prompts in the code.
    """
    return "\n".join([line.strip() for line in text.split("\n")])


def compute_cumulative_reward(all_rewards: List[List[float]], horizon: int):
    all_rewards = np.vstack(all_rewards)
    cum_rewards = np.cumsum(all_rewards, axis=1)

    all_rewards = cum_rewards / np.arange(1, horizon + 1)

    reward_means = np.mean(all_rewards, axis=0)
    reward_sems = scipy.stats.sem(all_rewards, axis=0)

    return reward_means, reward_sems


def plot_cumulative_reward(all_rewards: List[List[float]], horizon: int, title=None, filename=None):
    """
    Plot cumulative reward over time with confidence interval.
    
    Args:
        all_rewards: List of reward sequences [num_trials, horizon] 
        horizon: Number of timesteps
        title: Optional plot title
    """
    reward_means, reward_sems = compute_cumulative_reward(all_rewards, horizon)

    plt.figure(figsize=(8, 6))

    # Plot main line
    plt.plot(range(horizon), reward_means, color='#2ecc71', linewidth=2.5)

    # Add confidence interval
    plt.fill_between(
        range(horizon),
        reward_means - reward_sems,
        reward_means + reward_sems,
        alpha=0.2,
        color='#2ecc71'
    )

    # Add grid
    plt.grid(True, linestyle='--', alpha=0.7)

    # Customize axes
    plt.ylabel("Average Reward Over Trials")
    plt.xlabel("Number of Interactions (Horizon)")

    if title is not None:
        plt.title(title, pad=20)

    # Customize appearance    
    plt.tight_layout()
    if filename is not None:
        plt.savefig(filename, dpi=120, bbox_inches='tight')

    plt.show()


def plot_multi_cumulative_reward(config_name_to_all_rewards: Dict[str, List[List[float]]], horizon: int, title=None, filename=None):
    """
    Plot multiple cumulative reward curves with confidence intervals.
    
    Args:
        config_name_to_all_rewards: Dict mapping config names to reward sequences
        horizon: Number of timesteps
        title: Optional plot title
    """

    plt.figure(figsize=(8, 6))

    # Plot lines for each config
    for i, (config_name, all_rewards) in enumerate(config_name_to_all_rewards.items()):
        reward_means, reward_sems = compute_cumulative_reward(all_rewards, horizon)

        # Plot main line
        plt.plot(range(horizon), reward_means, label=config_name, linewidth=2.5)

        # Add confidence interval
        plt.fill_between(
            range(horizon),
            reward_means - reward_sems,
            reward_means + reward_sems,
            alpha=0.2,
        )

    # Add grid
    plt.grid(True, linestyle='--', alpha=0.7)

    # Customize axes
    plt.ylabel("Average Reward Over Trials")
    plt.xlabel("Number of Interactions (Horizon)")

    if title is not None:
        plt.title(title, fontweight='bold', pad=20)

    # Customize appearance
    plt.legend(loc='lower right')
    plt.tight_layout()
    if filename is not None:
        plt.savefig(filename, dpi=120, bbox_inches='tight')

    plt.show()

def calculate_cost(model, input_seq, output_seq):
    """
    Calculate the cost in dollars for a given sequence of tokens.

    Returns:
        dict: The input and output costs (in dollars).
    """
    return completion_cost(model=model, prompt=input_seq, completion=output_seq)


from banditbench.agents.guides import ActionInfoField, UCBGuide, ThompsonSamplingGuide
from banditbench.tasks.mab.env import BernoulliBandit
from banditbench.agents.classics import UCBAgent, ThompsonSamplingAgent

def test_action_info_field():
    print()
    print(str(ActionInfoField("exploration bonus", 0.003) + ActionInfoField("exploitation value", 0.003)))
    print(str(ActionInfoField("exploration bonus", "inf") + ActionInfoField("exploitation value", "inf")))

def test_action_info():
    field1 = ActionInfoField("filed 1", 1, "action info=(semantic meaning for field 1={:.2f}")
    field2 = ActionInfoField("filed 2", 2, "semantic meaning for field 2={:.2f}")
    field3 = ActionInfoField("filed 3", 3, "semantic meaning for field 3={:.2f})")
    # this gives us: action info=(semantic meaning for field 1=1.00, semantic meaning for field 2=2.00, semantic meaning for field 3=3.00)
    print(str(field1 + field2 + field3))

def test_ucb_guide():
    core_bandit = BernoulliBandit(2, 10, [0.5, 0.2], 123)
    ucb_agent = UCBAgent(core_bandit)
    ucb_agent.reset()

    ucb_guide = UCBGuide(ucb_agent)
    print(ucb_guide.get_action_guide_info(0).to_str())

def test_ts_guide():
    core_bandit = BernoulliBandit(2, 10, [0.5, 0.2], 123)
    ts_agent = ThompsonSamplingAgent(core_bandit)
    ts_agent.reset()

    ts_guide = ThompsonSamplingGuide(ts_agent)
    ts_guide = ThompsonSamplingGuide(core_bandit)
    ts_guide = ThompsonSamplingGuide(core_bandit, alpha_prior=2.0)
    print(ts_guide.get_action_guide_info(0).to_str())

test_action_info_field()
test_action_info()
test_ucb_guide()


import numpy as np
from banditbench.tasks.types import State
from banditbench.tasks.cb.movielens import MovieLens, MovieLensVerbal

def test_save_state():
    print()
    s = State(feature=[1, 2, 3], index=1, info=None)
    print(s.model_dump_json())
    s = State(feature=np.array([1, 2, 3]), index=1, info={'field1': 1})
    print(s.model_dump_json())
    s = State(feature=np.int32(3), index=1, info={'field1': np.array([1,2,3])})
    print(s.model_dump_json())

test_save_state()

env = None
verbal_env = None


def init_cb_env():
    global env
    global verbal_env

    if env is None:
        env = MovieLens('100k-ratings', num_arms=5, horizon=200, rank_k=5, mode='train',
                        save_data_dir='./tensorflow_datasets/')
    if verbal_env is None:
        verbal_env = MovieLensVerbal(env)


def test_action_name():
    init_cb_env()

    state, info = verbal_env.reset()
    new_state, reward, _, info = verbal_env.step(state, 'Star Wars (1977) (Action|Adventure|Romance|Sci-Fi|War)')
    assert info['interaction'].is_random is False



import numpy as np
from typing import Union, List, Optional
from banditbench.tasks.types import State

class BanditScenario:
    name: str

    action_names: List[str]
    action_unit: str

    base_description: str
    detailed_description: str
    query_prompt: str = ("\n\nWhich {unit} will you choose next? PLEASE RESPOND ONLY WITH {choices} AND NO TEXT "
                         "EXPLANATION.")

    fewshot_prompt: str = (
                "Here are some examples of optimal actions under different scenarios."
                " Use them as hints to help you come up with better actions.\n"
            )

    def __init__(self, num_actions: int,
                 action_names: List[str], action_unit: str,
                 base_description: str, detailed_description: str,
                 query_prompt: str,
                 seed: Union[int, None] = None):

        self.action_names = action_names
        self.action_unit = action_unit
        self.base_description = base_description
        self.detailed_description = detailed_description
        self.query_prompt = query_prompt

        if seed is not None:
            self.set_seed(seed)
            self.np_random.shuffle(self.action_names)
            self.action_names = self.action_names[:num_actions]
        else:
            self.action_names = self.action_names[:num_actions]

    def set_seed(self, seed: Optional[int] = None) -> None:
        self.seed = seed
        self.np_random = np.random.default_rng(self.seed)

    def get_instruction(self, version="base") -> str:
        raise NotImplementedError


class MABScenario(BanditScenario):
    def get_instruction(self, version="base"):
        """We add the few-shot examples in here"""
        if version == "base":
            prompt = self.base_description.format(
                len(self.action_names), "[" + ", ".join(self.action_names) + "]"
            )
        elif version == "detailed":
            prompt = self.detailed_description.format(
                len(self.action_names), "[" + ", ".join(self.action_names) + "]"
            )
        else:
            raise ValueError(f"Unknown description version {version}")

        return prompt

    def get_query_prompt(self) -> str:
        prompt = self.query_prompt.format(unit=self.action_unit, choices="[" + ", ".join(self.action_names) + "]")
        return prompt


class CBScenario(BanditScenario):
    def get_query_prompt(self, state: State, side_info: Optional[str] = None) -> str:
        """For contextual bandit, the agent can pass in optional side_info to the decision query context"""
        raise NotImplementedError



from typing import Dict, Any, Tuple, Union, List, Optional
import json
import numpy as np
from banditbench.tasks.scenario import BanditScenario
from banditbench.tasks.types import Action


class Bandit:
    num_arms: int
    horizon: int
    seed: Optional[int]
    h: int  # number of interactions

    def initialize_defaults(self) -> None:
        raise NotImplementedError

    def reset(self, seed: Optional[int] = None) -> Tuple[None, Dict[str, Any]]:
        raise NotImplementedError

    def set_seed(self, seed: Optional[int] = None) -> None:
        self.seed = seed
        self.np_random = np.random.default_rng(self.seed)

    @property
    def name(self) -> str:
        # b_vid_arms5_easy
        raise NotImplementedError


class VerbalBandit(Bandit):
    bandit_scenario: BanditScenario

    def __init__(self, core_bandit, *args, **kwargs):
        self.core_bandit = core_bandit
        self.num_arms = self.core_bandit.num_arms
        self.horizon = self.core_bandit.horizon
        self.seed = self.core_bandit.seed
        self.h = self.core_bandit.h

    def verbalize_feedback(self, action_name: Action, reward: float) -> str:
        """This corresponds to raw / unprocessed feedback from the environment"""
        raise NotImplementedError

    def verbalize_state(self, observation: Any) -> str:
        raise NotImplementedError

    @property
    def action_names(self) -> List[str]:
        raise NotImplementedError

    def get_query_prompt(self, *args, **kwargs) -> str:
        raise NotImplementedError

    def get_task_instruction(self, *args, **kwargs) -> str:
        raise NotImplementedError


# this is for serialization and for string translation



from banditbench.tasks.cb.movielens import MovieLens
from banditbench.tasks.mab import BernoulliBandit

def test_mab_traj():
    core_bandit = BernoulliBandit(4, 500, [0.25, 0.25, 0.25, 0.75])
    core_bandit.step(2)
    core_bandit.step(3)

    traj = core_bandit.history[0] + core_bandit.history[1]
    print(traj)
    print(traj.model_dump())
    print(traj.model_dump_json())

def test_cb_traj():
    env = MovieLens('100k-ratings', num_arms=5, horizon=200, rank_k=5, mode='train', save_data_dir='./tensorflow_datasets/')
    obs, _ = env.reset()
    env.step(obs, 3)
    env.step(obs, 4)
    traj = env.history[0] + env.history[1]
    print(traj)
    print(traj.model_dump())
    print(traj.model_dump_json())

test_mab_traj()
test_cb_traj()


import os
import setuptools

here = os.path.abspath(os.path.dirname(__file__))
version = {}
with open(os.path.join(here, "banditbench/version.py"), encoding="utf8") as fp:
    exec(fp.read(), version)
__version__ = version["__version__"]


install_requires = [
    "pydantic==2.10.5",
    "litellm",
    "pandas",
    "jinja2==3.1.2",
    "tensorflow_datasets==4.9.7",
    "tensorflow==2.15.0",
    "numpy==1.24.3",
    "matplotlib",
    "scipy"
]

setuptools.setup(
    name="banditbench",
    version=__version__,
    author="Allen Nie",
    author_email="anie@cs.stanford.edu",
    url="https://github.com/allenanie/EVOLvE",
    license='MIT LICENSE',
    description="BanditBench: A Bandit Benchmark to Evaluate Self-Improving LLM Algorithms",
    long_description=open('README.md', encoding="utf8").read(),
    packages=setuptools.find_packages(include=["banditbench*"]),
    install_requires=install_requires,
    python_requires=">=3.9",
)



from banditbench.tasks.mab import BernoulliBandit, VerbalMultiArmedBandit
from banditbench.agents.llm import LLMAgent


def test_async():
    try:
        core_bandit = BernoulliBandit(5, horizon=20, arm_params=[0.2, 0.2, 0.2, 0.2, 0.5])
        verbal_bandit = VerbalMultiArmedBandit(core_bandit, "ClothesShopping")
        agent = LLMAgent.build_with_env(verbal_bandit, summary=True, model="gpt-3.5-turbo")
        data = agent.in_context_learn(verbal_bandit, 5)
    except Exception as e:
        print(e)
        print("Omit the LLM Async test")


if __name__ == '__main__':
    test_async()



import json
import numpy as np
from pydantic import BaseModel, field_serializer
from typing import Dict, Any, Tuple, Union, List, Optional, Annotated

from banditbench.tasks.types import Action, InteractionBase, State, Info
from banditbench.tasks.env import Bandit



class Interaction(BaseModel, InteractionBase):
    state: State
    action: Action
    reward: float
    is_random: Union[bool, None] = None

    def __init__(self, state: State, action: Action, reward: float,
                 is_random: Union[bool, None] = None) -> None:
        super().__init__(state=state, action=action, reward=reward, is_random=is_random)


class VerbalInteraction(BaseModel, InteractionBase):
    state: State
    action: Action
    mapped_action: int  # index
    mapped_action_name: str
    reward: float
    feedback: str
    is_random: Union[bool, None] = None

    def __init__(self, state: State, action: Action, mapped_action: Action, mapped_action_name: Action,
                 reward: float, feedback: str, is_random: Union[bool, None] = None) -> None:
        super().__init__(state=state, action=action, mapped_action=mapped_action,
                         mapped_action_name=mapped_action_name,
                         reward=reward, feedback=feedback, is_random=is_random)


class ContextualBandit(Bandit):
    history: List[Interaction]

    def sample_state(self) -> State:
        """
        We sample a state from the state distribution
        """
        raise NotImplementedError

    def reward_fn(self, state: State, action: Action) -> float:
        """In a contextual bandit, this is a function f(x, a)"""
        raise NotImplementedError

    def reset(self, seed: Optional[int] = None) -> Tuple[State, Info]:
        raise NotImplementedError

    def step(self, state: State, action: Action) -> Tuple[State, float, bool, Dict[str, Any]]:
        raise NotImplementedError

    @property
    def verbal_info(self) -> Dict[str, Any]:
        """
        CB might be able to provide additional information from the dataset about the state
        This property is used by the VerbalBandit
        :return:
        """
        raise NotImplementedError

    @property
    def feature_dim(self) -> int:
        """
        :return: dimension of the contextual feature space0.

        """
        raise NotImplementedError
